'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const isNumber = subject => typeof subject === 'number';

const isArray = Array.isArray;

// Dynamic Weighted Moving Average

// @param {Number|Array.<Number>} alpha
var dma = (data, alpha, noHead) => {

  const length = data.length;

  if (alpha > 1) {
    return Array(length)
  }

  if (alpha === 1) {
    return data.slice()
  }

  const noArrayWeight = !isArray(alpha);
  const ret = [];

  let datum;

  // period `i`
  let i = 0;

  // `s` is the value of the DWMA at any time period `i`
  let s = 0;

  // Handles head
  for (; i < length; i ++) {
    datum = data[i];

    if (
      isNumber(datum)
      && (
        noArrayWeight
        || isNumber(datum)
      )
    ) {

      ret[i] = noHead
        ? 0
        : datum;

      s = datum;
      i ++;

      break
    }
  }

  // Dynamic weights: an array of weights
  // Ref:
  // https://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average
  // with a dynamic alpha
  if (!noArrayWeight) {
    for (; i < length; i ++) {
      datum = data[i];

      isNumber(datum) && isNumber(alpha[i])
        ? s =
          ret[i] = alpha[i] * datum + (1 - alpha[i]) * s
        : ret[i] = ret[i - 1];
    }

    return ret
  }

  const o = 1 - alpha;

  // Fixed alpha
  for (; i < length; i++) {
    datum = data[i];

    isNumber(datum)
      ? s =
        ret[i] = alpha * datum + o * s
      : ret[i] = ret[i - 1];
  }

  return ret
};

// Smoothed moving average

var sma = (data, size, times = 1) => dma(data, times / size, 1);

// Exponential moving average with 86% total weight

var ema = (data, size) => dma(data, 2 / (size + 1));

// simple moving average

var ma = (data, size) => {
  const length = data.length;

  if (!size) {
    return data.reduce((a, b) => a + b) / length
  }

  if (size <= 1) {
    return data.slice()
  }

  if (size > length) {
    return Array(length)
  }

  const prepare = size - 1;
  const ret = [];
  let sum = 0;
  let i = 0;
  let counter = 0;
  let datum;

  for (; i < length && counter < prepare; i ++) {
    datum = data[i];

    if (isNumber(datum)) {
      sum += datum;
      counter ++;
    }
  }

  for (; i < length; i ++) {
    datum = data[i];

    if (isNumber(datum)) {
      sum += datum;
    }

    if (isNumber(data[i - size])) {
      sum -= data[i - size];
    }

    ret[i] = sum / size;
  }

  return ret
};

// Weighted moving average

var wma = (data, size) => {
  const length = data.length;

  if (size <= 1) {
    return data.slice()
  }

  if (size > length) {
    return Array(length)
  }

  const ret = [];
  const denominator = size * (size + 1) / 2;
  const prepare = size - 1;
  let sum = 0;
  let numerator = 0;
  let datum = 0;
  let i = 0;
  let real = -1;


  for (; i < prepare; i ++) {
    datum = data[i];

    if (isNumber(datum)) {
      sum += datum;
      numerator += (i + 1) * datum;
    }
  }

  for (; i < length; i ++, real ++) {
    datum = data[i];

    if (isNumber(datum)) {
      sum += datum;
      numerator += size * datum;
    }

    if (real >= 0 && isNumber(data[real])) {
      sum -= data[real];
    }

    ret[i] = numerator / denominator;
    numerator -= sum;
  }

  return ret
};

const error = (
  message,
  // code
) => {
  const e = new Error(message);

  // if (code) {
  //   e.code = code
  // }

  throw e
};


const manipulate2Array = (a, b, mutator) => {
  if (a.length !== b.length) {
    error('the length of arrays not match');
  }

  return a.map((x, i) => mutator(x, b[i]))
};


const manipulateArray = (a, b, mutator) => {
  return a.map(x => mutator(x, b))
};


const isArray$1 = (a, b) => [a, b].map(Array.isArray);

const cleanArray = (array) => {
  array.forEach((item, i) => {
    if (item !== item) {
      delete array[i];
    }
  });
};

const orderUnaware = (
  a, b, mutator, mutatorReverse,
  ensureNumber
) => {
  const [A, B] = isArray$1(a, b);

  const ret = A
    ? B
      ? manipulate2Array(a, b, mutator)
      : manipulateArray(a, b, mutator)
    : B
      ? manipulateArray(b, a, mutatorReverse)
      : error('at least one array is required');

  if (ensureNumber) {
    cleanArray(ret);
  }

  return ret
};


const orderAware = (
  a, b, mutator,
  ensureNumber
) => {
  const [A, B] = isArray$1(a, b);

  const ret = A
    ? B
      ? manipulate2Array(a, b, mutator)
      : manipulateArray(a, b, mutator)
    : error('the first argument must be an array');

  if (ensureNumber) {
    cleanArray(ret);
  }

  return ret
};

const add = (a, b) => a + b;
const addReverse = (a, b) => b + a;

var add$1 = (a, b, n) =>
  orderUnaware(a, b, add, addReverse, n);

const sub = (a, b) => a - b;

var sub$1 = (a, b, n) => orderAware(a, b, sub, n);

const mul = (a, b) => a * b;

var mul$1 = (a, b, n) => orderUnaware(a, b, mul, mul, n);

const div = (a, b) => {
  if (b === 0) {
    error('divide by zero');
  }

  return a / b
};

var div$1 = (a, b, n) => orderAware(a, b, div, n);

var index = (
  data,
  slowPeriods = 26,
  fastPeriods = 12,
  signalPeriods = 9
) => {
  const MACD = sub$1(
    ema(data, fastPeriods),
    ema(data, slowPeriods),
    1
  );

  const signal = ema(MACD, signalPeriods);
  const histogram = mul$1(2, sub$1(MACD, signal), 1);

  return {
    MACD,
    signal,
    histogram
  }
};

const compare = (data, size, comparer) => {
  const length = data.length;

  if (size > length) {
    return Array(length)
  }

  if (!size) {
    return data.reduce(comparer)
  }

  if (size <= 1) {
    return data.slice()
  }

  let i = size - 1;
  const ret = [];

  for (; i < length; i ++) {
    ret[i] = reduce(data, i - size + 1, i + 1, comparer);
  }

  return ret
};


// @param {Number} start index to start [
// @param {Number} end index less than )
// -> [start, end)
const reduce = (array, start, end, reducer) => {
  let prev;
  let i = start;

  for (; i < end; i ++) if (i in array) {
    if (prev === undefined) {
      prev = array[i];
      continue
    }
    
    prev = reducer(prev, array[i]);
  }

  return prev
};


const hhv = (data, size) =>
  compare(data, size, (a, b) => Math.max(a, b));

const llv = (data, size) =>
  compare(data, size, (a, b) => Math.min(a, b));

const UPPER = 100;
const LOWER = 0;

const clean = n => Math.min(UPPER, Math.max(LOWER, n));

var index$1 = (
  closePrices,
  lowPrices,
  highPrices,
  periods = 9,
  kPeriods = 3,
  dPeriods = 3,
  kTimes = 3,
  dTimes = 2
) => {

  const lowest = llv(lowPrices, periods);
  const highest = hhv(highPrices, periods);

  const u = div$1(
    mul$1(sub$1(closePrices, lowest, 1), UPPER, 1),
    sub$1(highest, lowest, 1)
  );

  const ks = sma(u, kPeriods);
  const ds = sma(ks, dPeriods);
  const js = sub$1(
    mul$1(kTimes, ks, 1),
    mul$1(dTimes, ds, 1),
    1
  );

  return {
    K: ks.map(clean),
    D: ds.map(clean),
    J: js.map(clean)
  }
};

var deviation = (data, size) => {
  const length = data.length;
  const avg = ma(data, size);
  const ret = [];

  let i = size - 1;
  let j;
  let sum;

  for (; i < length; i ++) {
    sum = 0;
    j = i - size + 1;

    for (; j <= i; j ++) {
      sum += Math.pow(data[j] - avg[i], 2);
    }

    ret[i] = Math.sqrt(sum / size);
  }

  return ret
};

var index$2 = (datum, size = 20, times = 2, {
  ma: avg,
  sd
} = {}) => {

  avg = avg || ma(datum, size);
  sd = sd || deviation(datum, size);

  const timesSd = mul$1(sd, times);

  return {
    upper: add$1(avg, timesSd),
    mid: avg,
    lower: sub$1(avg, timesSd)
  }
};

exports.macd = index;
exports.kdj = index$1;
exports.ma = ma;
exports.sma = sma;
exports.dma = dma;
exports.wma = wma;
exports.ema = ema;
exports.hhv = hhv;
exports.llv = llv;
exports.sd = deviation;
exports.boll = index$2;
exports.add = add$1;
exports.sub = sub$1;
exports.mul = mul$1;
exports.div = div$1;
